//============================================================================
// Sample LCD drawing routines for displays with horizontally aligned data,
// and the MSB to the LEFT. This was degeloped on a Sitronix ST7920 based
// Crystalfontz CFAG14432D-TFH-TT 144x32 display :
//
// https://www.crystalfontz.com/search.php?q=cfag14432d
//
//---------------------------------------------------------------------------
// Written by Brent A. Crosby, Crystalfontz America, Inc.
// http://www.crystalfontz.com
//---------------------------------------------------------------------------
//This is free and unencumbered software released into the public domain.
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non-commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain. We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors. We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>
//============================================================================
#include "Arduino.h"
#include "typedefs.h"
#include "fnt_08x08_h.h"
#include "fnt_12x16_h.h"
#include "lcd_draw.h"
//============================================================================
// The addressing of the ST7920 in graphics mode is particularly nasty. You
// can only address a given horizontal group of 16 pixels, then you must
// write two bytes to access the right half of that 16-pixel group.
// To get around this messiness, we will keep a frame buffer (576 bytes) in
// our local RAM that is byte adressible and can be written and read.
#define FRAME_WIDTH_BYTES (18)
#define FRAME_WIDTH_PIXELS (FRAME_WIDTH_BYTES*8)
#define FRAME_HEIGHT_PIXELS (32)
uint8_t framebuffer[FRAME_HEIGHT_PIXELS][FRAME_WIDTH_BYTES];
//============================================================================
void Fill_Screen(uint8_t fill_data)
  {
  //Clear the display memory
  memset(framebuffer,fill_data,sizeof(framebuffer));
  }
//============================================================================
// put_letter() will display a single character from the 08x08 font at (x,y)
// It will pad a 1-pixel border to the left and top of the letter. The 08x08
// font already has a 1-pixel border at the right and bottom. This means that
// a 9x9 block will be modified.
//
// put_letter will use a "COPY_PUT"--replacing the area occupied by the
// character. This means that you can call put_letter() several times and
// then call UpdateLCD() without having to clear rectangles.
//---------------------------------------------------------------------------
void put_letter(uint8_t x, uint8_t y, uint8_t letter)
  {
  register uint8_t
    line;
  register uint8_t
    low_3_bits_of_x;
  register WORD_UNION
    *Pixel_Data;
  register const uint8_t
    *Font_Pointer;
  register uint16_t
    Clearing_Mask;

  //Bail for garbage in. We need at least 9 pixels below and to the right.
  if(((FRAME_WIDTH_PIXELS-9)<x)||
     ((FRAME_HEIGHT_PIXELS-9)<y))
    return;

  //Since we need to fuss with 9 bits in width, we will always have to
  //look at two bytes of display memory per line.

  //Calculate the address of the first byte in display in LCD_Memory,
  //and the second byte by the union.
  Pixel_Data=(WORD_UNION *)&framebuffer[y][x>>3];

  //Get a pointer into the font information for this character.
  Font_Pointer=&Font_08x08_Horiz[(letter-FONT_08X08_HORIZ_BASE)][0];

  //Avoid doing this mask a bunch of times.
  low_3_bits_of_x=x&0x07;

  //Avoid doing this shift a bunch of times
  Clearing_Mask=(0x7F00)>>low_3_bits_of_x;

  for(line=0;line<=8;line++)
    {
    //Clear the 9 bits in the display memory
    Pixel_Data->as_word&=Clearing_Mask;

    //0r in the data from the font
    if(line) //top line is blank
      {
      Pixel_Data->as_bytes[0]|=
        ((pgm_read_byte(Font_Pointer))>>((low_3_bits_of_x)+1));
      Pixel_Data->as_bytes[1]|=
        ((pgm_read_byte(Font_Pointer))<<(7-(low_3_bits_of_x)));
      //Move to the next line in the font
      Font_Pointer++;
      }

    //Move to the next line in the display. Pixel_Data is a WORD_UNION (2
    //bytes) so we only need to 1/2 of the framebuffer width.
    Pixel_Data+=(FRAME_WIDTH_BYTES/2);
    }
  }
//============================================================================
// void put_string(uint8_t x, uint8_t y, uint8_t Field_Width, sbyte flash *input);
//
// put_string() will display the null-terminated string pointed to by input
// at (x,y) in the 08x08 font. It will pad a 1-pixel border to the left and
// top of the text. The 08x08 font already has a 1-pixel border at the
// right and bottom.
//
// Put_String will use a "COPY_PUT", replacing the area occupied by the
// string. This means that you can call put_string() several times and then call
// UpdateLCD(), without having to clear rectangles.
//---------------------------------------------------------------------------
void put_string_P(uint8_t x, uint8_t y, uint8_t Field_Width,const char *input)
  {
  register uint8_t
    Terminator_Found;
  register uint8_t
    Characters_Placed;
  register uint8_t
    this_character;

  //Initialize Terminator_Found.
  Terminator_Found=0;
  //Move across the field. We will either put the character or a blank
  //in every position of Field_Width.
  for(Characters_Placed=0;Characters_Placed<Field_Width;Characters_Placed++)
    {
    //If we have not passed the terminator, then get the next
    //character in the string. If we find the terminator,
    //remember that we are out of characters.
    if(!Terminator_Found)
      {
      this_character=pgm_read_byte(input++);
      if(!this_character)
        {
        Terminator_Found=1;
        this_character=' ';
        }
      }
    else
      this_character=' ';

    //Print the character. put_letter does clipping.
    put_letter(x,y,this_character);

    //Move over one character.
    x+=8;
    }
  }
//---------------------------------------------------------------------------
void put_string(uint8_t x, uint8_t y, uint8_t Field_Width, char *input)
  {
  register uint8_t
    Terminator_Found;
  register uint8_t
    Characters_Placed;
  register uint8_t
    this_character;

  //Initialize Terminator_Found.
  Terminator_Found=0;
  //Move across the field. We will either put the character or a blank
  //in every position of Field_Width.
  for(Characters_Placed=0;Characters_Placed<Field_Width;Characters_Placed++)
    {
    //If we have not passed the terminator, then get the next
    //character in the string. If we find the terminator,
    //remember that we are out of characters.
    if(!Terminator_Found)
      {
      this_character=*input++;
      if(!this_character)
        {
        Terminator_Found=1;
        this_character=' ';
        }
      }
    else
      this_character=' ';

    //Print the character. put_letter does clipping.
    put_letter(x,y,this_character);

    //Move over one character.
    x+=8;
    }
  }
//============================================================================
void horizontal_line(uint8_t x1, uint8_t y, uint8_t x2)
  {
  register uint8_t
    *LCD_Memory_1;
  register uint8_t
    *LCD_Memory_2;
  register uint8_t
    mask_left;
  register uint8_t
    mask_right;

  //Bail for bogus parametrers.
  if((x2<x1)||
     (FRAME_HEIGHT_PIXELS<=y)||
     (FRAME_WIDTH_PIXELS<=x1)||
     (FRAME_WIDTH_PIXELS<=x2))
    return;
  //Draw the last pixel too.
  x2++;

  //Calculate the ends.
  mask_left=  (0xFF>>(x1&0x07));
  mask_right=~(0xFF>>(x2&0x07));

  //Calculate the address of the first byte in display in LCD_Memory
  LCD_Memory_1=&framebuffer[y][x1>>3];
  //Calculate the address of the last byte in display in LCD_Memory
  LCD_Memory_2=&framebuffer[y][x2>>3];
  //If they are the same, then this is a special case.
  if(LCD_Memory_1==LCD_Memory_2)
    //The line fits in one byte.
    *LCD_Memory_1|=mask_left&mask_right;
  else
    {
    //This is the case where we need to loop over multiple
    //(possibly 0) colums. Do the left byte.
    *LCD_Memory_1|=mask_left;
    //Do the bytes in the middle.
    for(LCD_Memory_1++;LCD_Memory_1!=LCD_Memory_2;LCD_Memory_1++)
      *LCD_Memory_1=0xFF;
    //Do the right byte.
    *LCD_Memory_2|=mask_right;
    }
  }
//============================================================================
void vertical_line(uint8_t x, uint8_t y1, uint8_t y2)
  {
  register uint8_t
    *LCD_Memory;
  register uint8_t
    row;
  register uint8_t
    Set_Mask;

  //Bail for bogus parametrers.
  if((y2<y1)||
     (FRAME_WIDTH_PIXELS<=x)||
     (FRAME_HEIGHT_PIXELS<=y1)||
     (FRAME_HEIGHT_PIXELS<=y2))
    return;

  //Calculate the address of the first byte in display in LCD_Memory
  LCD_Memory=&framebuffer[y1][x>>3];

  //Calculate Set_Mask, the horizontal mask that we will or with
  //LCD_Memory to set the pixel.
  Set_Mask=0x80>>(x&0x07);

  //Move down memory, oring the mask in.
  for(row=y1;row<=y2;row++)
    {
    *LCD_Memory|=Set_Mask;
    LCD_Memory+=FRAME_WIDTH_BYTES;
    }
  }
//============================================================================
void outline_rectangle(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2)
  {
  horizontal_line(x1+1,y1,x2-1);
  horizontal_line(x1+1,y2,x2-1);
  vertical_line(x1,y1,y2);
  vertical_line(x2,y1,y2);
  }
//============================================================================
void clear_rectangle(uint8_t x1,uint8_t y1,uint8_t x2,uint8_t y2)
  {
  register uint8_t
    mask_left;
  register uint8_t
    mask_right;
  register uint8_t
    *LCD_Memory;
  register uint8_t
    *LCD_Memory_Left;
  register uint8_t
    *LCD_Memory_Right;
  register uint8_t
    row;

  //Bail for bogus parametrers. 
  if((x2<x1)||
     (y2<y1)||
     (FRAME_WIDTH_PIXELS<=x1)||
     (FRAME_WIDTH_PIXELS<=x2)||
     (FRAME_HEIGHT_PIXELS<=y1)||
     (FRAME_HEIGHT_PIXELS<=y2))
    return;

  //Draw the last pixel too.
  x2++;

  //Calculate the address of the left byte in display in LCD_Memory
  LCD_Memory_Left=&framebuffer[y1][x1>>3];
  //Calculate the address of the right byte in display in LCD_Memory
  LCD_Memory_Right=&framebuffer[y1][x2>>3];

  //Calculate the ends outside the loop.
  mask_left=~(0xFF>>(x1&0x07));
  mask_right=(0xFF>>(x2&0x07));

  //If they are the same, then this is a special case.
  if(LCD_Memory_Left==LCD_Memory_Right)
    {
    //The rectangle fits in one column of display memory. Calculate a
    //new clearing mask that is common between the right and left.
    mask_left=mask_right|mask_left;

    for(row=y1;row<=y2;row++)
      {
      *LCD_Memory_Left&=mask_left;
      LCD_Memory_Left+=FRAME_WIDTH_BYTES;
      }
    }
  else
    {
    //Loop over the rows.
    for(row=y1;row<=y2;row++)
      {
      LCD_Memory=LCD_Memory_Left;
      //This is the case where we need to loop over multiple
      //(possibly 0) columns. Do the left byte.
      *LCD_Memory&=mask_left;
      //Do the bytes in the middle.
      for(LCD_Memory++;LCD_Memory!=LCD_Memory_Right;LCD_Memory++)
        *LCD_Memory=0x00;
      //Do the right byte.
      *LCD_Memory_Right&=mask_right;
      //Move to the next row
      LCD_Memory_Left+=FRAME_WIDTH_BYTES;
      LCD_Memory_Right+=FRAME_WIDTH_BYTES;
      }
    }
  }
//============================================================================
void Draw_Bitmap(uint8_t x,
                 uint8_t y,
                 uint8_t x_size,
                 uint8_t y_size,
                 const uint8_t *Bitmap_Base_Pointer)
  {
  register WORD_UNION
    Bitmap_Data;
  register uint8_t
    low_three_bits_of_x;
  register uint8_t
    *LCD_Memory;
  register uint8_t
    *LCD_Memory_Right;
  register uint8_t
    *LCD_Memory_Left;
  register uint8_t
    mask_left;
  register uint8_t
    mask_right;
  register uint8_t
    row;
  register const uint8_t
    *Bitmap_Pointer;
  register uint8_t
    Bitmap_Line_Increment;

  uint8_t
    x2;
  uint8_t
    low_three_bits_of_x2;

  //Bail for garbage in.
  if(((FRAME_WIDTH_PIXELS-x_size)<x)||
     ((FRAME_HEIGHT_PIXELS-y_size)<y))
    return;

  //Calculate the address of the upper left byte in LCD_Memory
  LCD_Memory_Left=&framebuffer[y][x>>3];
  //Calculate the address of the upper right byte in LCD_Memory
  x2=x+x_size;
  LCD_Memory_Right=&framebuffer[y][x2>>3];

  //Used a lot, Calculate once outside the loops.
  low_three_bits_of_x=x&0x07;
  low_three_bits_of_x2=x2&0x07;

  //Calculate how far to skip forward in the bitmap every line.
  Bitmap_Line_Increment=(x_size>>3);
  if(x_size&0x07)
    Bitmap_Line_Increment++;

  //Calculate the ends outside the loops.
  mask_left=~(0xFF>>(low_three_bits_of_x));
  mask_right=(0xFF>>(low_three_bits_of_x2));

  //If they are the same, then this is a special case.
  if(LCD_Memory_Left==LCD_Memory_Right)
    {
    //Set the bitmap pointer to the start of this line
    Bitmap_Pointer=Bitmap_Base_Pointer;
    //The bit-blt fits in one column of display memory. Calculate a
    //new clearing mask that is common between the right and left.
    mask_left=mask_right|mask_left;

    for(row=0;row<y_size;row++)
      {
      //Clear the pixels.
      *LCD_Memory_Left&=mask_left;
      //Or in the Bitmap data, and move to the next line.
      //The bitmap data can only be one byte wide, since it does
      //not cross a byte boundary in display memory.
      *LCD_Memory_Left|=
        (pgm_read_byte(Bitmap_Pointer++)>>(low_three_bits_of_x))&(~mask_left);
      //Move to the next line in display memory.
      LCD_Memory_Left+=30;
      }
    }
  else
    {
    //This is the case where the bitmap spans at least two (or more) columns
    //of display memory.
    //Loop over the rows.
    for(row=0;row<y_size;row++)
      {
      //Set the bitmap pointer to the start of this line
      Bitmap_Pointer=Bitmap_Base_Pointer;
      //Set the display pointer to the start of this line.
      LCD_Memory=LCD_Memory_Left;
      //Do the left-most column, clear first then set.
      *LCD_Memory&=mask_left;
      *LCD_Memory++|=pgm_read_byte(Bitmap_Pointer)>>(low_three_bits_of_x);

      //Now we need to loop over multiple (possibly 0) 8-bit wide columns.
      while(LCD_Memory!=LCD_Memory_Right)
        {
        //Need to reverse "endian" to get word shift to work correctly.
        Bitmap_Data.as_bytes[1]=pgm_read_byte(Bitmap_Pointer++);
        Bitmap_Data.as_bytes[0]=pgm_read_byte(Bitmap_Pointer);
        //Slide it over as a word
        Bitmap_Data.as_word>>=(low_three_bits_of_x);
        //Push it into the display memory.
        *LCD_Memory++=Bitmap_Data.as_bytes[0];
        }

      //Do the right byte. Clear the pixels first.
      *LCD_Memory_Right&=mask_right;
      //Now calculate which bits should be set.

      //Need to reverse "endian" to get word shift to work correctly.
      Bitmap_Data.as_bytes[1]=pgm_read_byte(Bitmap_Pointer++);
      Bitmap_Data.as_bytes[0]=pgm_read_byte(Bitmap_Pointer);
      //Slide it over as a word
      Bitmap_Data.as_word>>=(low_three_bits_of_x);
      //Put that into display memory, observing the mask.
      *LCD_Memory_Right|=Bitmap_Data.as_bytes[0]&(~mask_right);

      //Move to the next row of display memory
      LCD_Memory_Left+=FRAME_WIDTH_BYTES;
      LCD_Memory_Right+=FRAME_WIDTH_BYTES;
      //Move to the next line in the bitmap
      Bitmap_Base_Pointer+=Bitmap_Line_Increment;
      }
    }
  }
//============================================================================
// Only '0'-'9', '.', '+', '-', and ' ' will be displayed. The font is
// 12 x 16.
void put_big_number(uint8_t x, uint8_t y, uint8_t Field_Width, char *input)
  {
  register uint8_t
    Terminator_Found;
  register uint8_t
    Characters_Placed;
  register uint8_t
    this_character;

  //Initialize Terminator_Found.
  Terminator_Found=0;
  //Move across the field. We will either put the character or a blank
  //in every position of Field_Width.
  for(Characters_Placed=0;Characters_Placed<Field_Width;Characters_Placed++)
    {
    //If we have not passed the terminator, then get the next
    //character in the string. If we find the terminator,
    //remember that we are out of characters.
    if(!Terminator_Found)
      {
      this_character=*input++;
      if(!this_character)
        {
        Terminator_Found=1;
        this_character=' ';
        }
      }
    else
      this_character=' ';

    //Print the character. Draw_Bitmap does clipping.
    switch(this_character)
      {
      case '+': Draw_Bitmap(x,y,12,16,big_plus); break;
      case '-': Draw_Bitmap(x,y,12,16,big_minus); break;
      case '.': Draw_Bitmap(x,y,12,16,big_dot); break;
      case '0': Draw_Bitmap(x,y,12,16,big_0); break;
      case '1': Draw_Bitmap(x,y,12,16,big_1); break;
      case '2': Draw_Bitmap(x,y,12,16,big_2); break;
      case '3': Draw_Bitmap(x,y,12,16,big_3); break;
      case '4': Draw_Bitmap(x,y,12,16,big_4); break;
      case '5': Draw_Bitmap(x,y,12,16,big_5); break;
      case '6': Draw_Bitmap(x,y,12,16,big_6); break;
      case '7': Draw_Bitmap(x,y,12,16,big_7); break;
      case '8': Draw_Bitmap(x,y,12,16,big_8); break;
      case '9': Draw_Bitmap(x,y,12,16,big_9); break;
      default: Draw_Bitmap(x,y,12,16,big_space);
      }

    //Move over one character.
    x+=12;
    }
  }
//============================================================================
void dec_3_to_string(uint8_t input, char *output)
  {
  uint8_t
    digit;
  uint8_t
    no_blank;
  digit=input/100;
  if(digit)
    {
    output[0]=digit+'0';
    no_blank=1;
    }
  else
    {
    output[0]=' ';
    no_blank=0;
    }
  input%=100;
  digit=input/10;
  if(digit|no_blank)
    {
    output[1]=digit+'0';
    }
  else
    {
    output[1]=' ';
    }
  output[2]=(input%10+'0');
  output[3]=0;
  }
//============================================================================

